#define _GNU_SOURCE
#include <stddef.h>
#include <sys/types.h>
#include <termios.h>
#include <unistd.h>

#define SERVICE "login: "
#define FATAL SERVICE "fatal: "
#define WARNING SERVICE "warning: "

int write_buf(int const fd, char const* buf, size_t len) {
  while (len!=0) {
    ssize_t const nwrit=write(fd, buf, len);
    if (nwrit<0) {
      if (errno==error_intr)
        continue;
      return -1;
    }
    buf+=nwrit;
    len-=nwrit;
  }
  return 0;
}

int read_line(int const fd, char* buf, size_t size, size_t* const plen) {
  size_t len=0;
  while (size!=0) {
    ssize_t const nread=read(fd, buf, 1);
    if (nread<0) {
      if (errno==error_intr)
        continue;
      return -1;
    }
    if (nread==0)
      break;
    --size;
    ++len;
    if (*(buf++)=='\n')
      break;
  }
  *plen=len;
  return 0;
}

void check(char* const buf, size_t const len, size_t const max,
          char const* const name) {
  char const* err=0;
  if (len==0 || buf[len-1]!='\n')
    strerr_die3x(111, FATAL, "EOF while reading ", (char*)name);
  if (len==max)
    err=" too long\n";
  else if (byte_chr(buf, len, '\0')!=len)
    err=" contains null characters\n";
  buf[len-1]='\0';
  if (err==0)
    return;
  buffer_puts(buffer_1, (char*)name);
  buffer_puts(buffer_1, (char*)err);
  buffer_flush(buffer_1);
  buffer_puts(buffer_2, FATAL);
  buffer_puts(buffer_2, (char*)name);
  buffer_puts(buffer_2, (char*)err);
  buffer_flush(buffer_2);
  _exit(111);
}

int main(int argc, char** argv) {
  char* x;
  typedef char buf[512];
  static size_t const bufmax=(sizeof (buf))-1;
  buf ubuf, pbuf;
  size_t ulen, plen;
  struct termios normal, noecho;
#ifdef TCSASOFT
  int const tcsa=TCSAFLUSH|TCSASOFT;
#else
  int const tcsa=TCSAFLUSH;
#endif
  int ret, err;
  char const* const cvm=env_get("CVM_PLAIN");
  struct stralloc sa={ 0 };

  if (cvm==0)
    strerr_die2x(111, FATAL, "$CVM_PLAIN not set");
  if (tcgetattr(0, &normal)!=0)
    strerr_die2sys(111, FATAL, "unable to get terminal attributes: ");
  noecho=normal;
  noecho.c_lflag&=~(ECHO|ISIG);

  strerr_warn3(SERVICE, "tty ", env_get("TTY"), 0);

  x=env_get("PROMPT_LOGIN");
  if (x==0)
    x="login: ";
  if (write_buf(1, x, str_len(x))!=0)
    strerr_die2sys(111, FATAL, "unable to write login prompt: ");
  if (read_line(0, ubuf, bufmax, &ulen)!=0)
    strerr_die2sys(111, FATAL, "unable to read username: ");
  check(ubuf, ulen, bufmax, "username");

  strerr_warn3(SERVICE, "attempt ", ubuf, 0);

  if (env_get("DENY")!=0) {
    if (stralloc_copys(&sa, "ALLOW_")==0 ||
        stralloc_cats(&sa, ubuf)==0 ||
        stralloc_0(&sa)==0)
      strerr_die1sys(111, FATAL);
    err=(env_get(sa.s)==0);
  } else {
    if (stralloc_copys(&sa, "DENY_")==0 ||
        stralloc_cats(&sa, ubuf)==0 ||
        stralloc_0(&sa)==0)
      strerr_die1sys(111, FATAL);
    err=(env_get(sa.s)!=0);
  }
  if (err) {
    x="login denied";
    buffer_puts(buffer_1, x);
    buffer_puts(buffer_1, "\n");
    buffer_flush(buffer_1);
    strerr_die2x(100, SERVICE, x);
  }

  x=env_get("PROMPT_PASSWORD");
  if (x==0)
    x="Password: ";
  if (write_buf(1, x, str_len(x))!=0)
    strerr_die2sys(111, FATAL, "unable to write password prompt: ");
  if (tcsetattr(0, tcsa, &noecho)!=0)
    strerr_die2sys(111, FATAL, "unable to set terminal attributes: ");
  ret=read_line(0, pbuf, bufmax, &plen);
  err=errno;
  if (tcsetattr(0, tcsa, &normal)!=0)
    strerr_die2sys(111, FATAL, "unable to set terminal attributes: ");
  if (ret!=0) {
    errno=err;
    strerr_die2sys(111, FATAL, "unable to read password: ");
  }
  if (write_buf(1, "\n", 1)!=0)
    strerr_die2sys(111, FATAL, "unable to write newline: ");
  check(pbuf, plen, bufmax, "password");

  {
    char const* credentials[3]={ ubuf, pbuf, 0 };
    ret=authenticate(cvm, credentials);
  }
  if (ret!=0) {
    if (ret==CVME_PERMFAIL) {
      x="login incorrect";
      buffer_puts(buffer_1, x);
      buffer_puts(buffer_1, "\n");
      buffer_flush(buffer_1);
      strerr_die2x(100, SERVICE, x);
    }
    switch (ret) {
      case CVME_GENERAL:
        x="unspecified error";
      case CVME_BAD_CLIDATA:
        x="invalid data from client";
      case CVME_BAD_MODDATA:
        x="invalid data from module";
      case CVME_IO:
        x="I/O error";
      default:
        x="unknown error";
    }
    buffer_puts(buffer_1, "error authenticating\n");
    buffer_flush(buffer_1);
    strerr_die2x(111, FATAL, x);
  }

  strerr_warn2(SERVICE, "login ok", 0);

  {
    pid_t const pid=fork();
    switch (pid) {
      case -1:
        strerr_die2sys(111, FATAL, "unable to fork: ");
      case 0:
        break;
      default:
        if (tcsetpgrp(0, pid))
          strerr_die2sys(111, FATAL,
                         "unable to set foreground process group: ");
        if (wait_pid(0, pid)<0)
          strerr_warn2(WARNING, "unable to wait for child: ", &strerr_sys);
        strerr_warn2(SERVICE, "logout", 0);
        _exit(0);
    }
  }

  pathexec_env("USER", (char*)fact_username);
  if (stralloc_copyb(&sa, "", 0)==0)
    strerr_die1sys(111, FATAL);
  {
    char const* c;
    if (fact_str(FACT_USERID, &c)!=0)
      pathexec_env("UID", (char*)c);
    if (fact_str(FACT_GROUPID, &c)!=0) {
      pathexec_env("GID", (char*)c);
      if (env_get("INCLUDE_GID")!=0)
        if (stralloc_copys(&sa, " ")==0 ||
            stralloc_cats(&sa, (char*)c)==0)
          strerr_die1sys(111, FATAL);
    }
    if (fact_str(FACT_SUPP_GROUPID, &c)!=0)
      if (stralloc_cats(&sa, " ")==0 ||
          stralloc_cats(&sa, (char*)c)==0)
        strerr_die1sys(111, FATAL);
  }
  x=env_get("EXTRA_GROUPS");
  if (x!=0 && stralloc_cats(&sa, x)==0)
    strerr_die1sys(111, FATAL);
  if (stralloc_0(&sa)==0)
    strerr_die1sys(111, FATAL);
  pathexec_env("GROUPS", sa.s);
  pathexec_env("NAME", (char*)fact_realname);
  pathexec_env("HOME", (char*)fact_directory);
  pathexec_env("SHELL", (char*)fact_shell);

  if (setpgid(0, 0)!=0)
    strerr_die2sys(111, FATAL, "unable to set process group: ");
  if (tcsetpgrp(0, getpgid(0)))
    strerr_die2sys(111, FATAL, "unable to set foreground process group: ");
  close(2);
  if (dup(1)!=2)
    _exit(111);

  {
    char* usercmd[]={ package_home "/bin/setstate", "tgGuh-", 0 };
    if (argc>1)
      ++argv;
    else
      argv=usercmd;
    pathexec(argv);
    strerr_die4sys(111, FATAL, "unable to execute ", argv[0], ": ");
  }

  return 111;
}
