#define _GNU_SOURCE
#include <sys/types.h>
#include <unistd.h>
#include <grp.h>

#define PROG "setstate"
#define FATAL PROG ": fatal: "
#define WARNING PROG ": warning: "

int main(int argc, char** argv) {
  char* s;
  unsigned int i;
  unsigned long l;
  int do_tty=0;
  int do_chroot=0;
  int do_setgid=0;
  int do_setgroups=0;
  int do_setuid=0;
  int do_home=0;
  int do_dash=0;

  if (argc>1) {
    s=argv[1];
    ++argv;
    --argc;
  } else
    s="";

  while (*s!='\0') {
    switch (*s) {
      case 't': do_tty=1;       break;
      case 'r': do_chroot=1;    break;
      case 'g': do_setgid=1;    break;
      case 'G': do_setgroups=1; break;
      case 'u': do_setuid=1;    break;
      case 'h': do_home=1;      break;
      case '-': do_dash=1;      break;
      default: {
        char opt[2]={ 0, 0 };
        *opt=*s;
        strerr_die3x(111, FATAL, "unrecognized option: ", opt);
      }
    }
    ++s;
  }

  if (do_tty!=0) {
    s=env_get("UID");
    if (s==0)
      strerr_die2x(111, FATAL, "$UID not set");
    i=scan_ulong(s, &l);
    if (i==0 || s[i]!='\0')
      strerr_die2x(111, FATAL, "$UID not numeric");
    if (l!=(uid_t)l)
      strerr_die2x(111, FATAL, "$UID overflows uid_t");
    s=env_get("TTY");
    if (s==0)
      strerr_die2x(111, FATAL, "$TTY not set");
    if (chown(s, l, -1)!=0)
      strerr_die4sys(111, FATAL, "unable to chown ", s, ": ");
  }

  if (do_chroot!=0) {
    s=env_get("ROOT");
    if (s==0)
      strerr_die2x(111, FATAL, "$ROOT not set");
    if (chdir(s)!=0)
      strerr_die4sys(111, FATAL, "unable to chdir to ", s, ": ");
    if (chroot(".")!=0)
      strerr_die4sys(111, FATAL, "unable to chroot to ", s, ": ");
  }

  if (do_setgid!=0) {
    s=env_get("GID");
    if (s==0)
      strerr_die2x(111, FATAL, "$GID not set");
    i=scan_ulong(s, &l);
    if (i==0 || s[i]!='\0')
      strerr_die2x(111, FATAL, "$GID not numeric");
    if (l!=(gid_t)l)
      strerr_die2x(111, FATAL, "$GID overflows gid_t");
    if (setgid(l)!=0)
      strerr_die4sys(111, FATAL, "unable to setgid to ", s, ": ");
  }

  if (do_setgroups!=0) {
    size_t ngroups;
    gid_t gid, * groups;
    char* t;
    static char const toolong[]="$GROUPS too long";
    static char const malformed[]="$GROUPS malformed";
    s=env_get("GROUPS");
    if (s==0)
      strerr_die2x(111, FATAL, "$GROUPS not set");
    t=s;
    ngroups=0;
    while (*t==' ') {
      ++ngroups;
      if (ngroups==0)
        strerr_die2x(111, FATAL, (char*)toolong);
      ++t;
      t+=str_chr(t, ' ');
    }
    i=ngroups*sizeof (gid_t);
    if (i/sizeof (gid_t)!=ngroups)
      strerr_die2x(111, FATAL, (char*)toolong);
    groups=(gid_t*)alloc(i);
    if (groups==0)
      strerr_die1sys(111, FATAL);
    ngroups=0;
    while (*s!='\0') {
      if (*s!=' ')
        strerr_die2x(111, FATAL, (char*)malformed);
      ++s;
      i=scan_ulong(s, &l);
      s+=i;
      if (i==0)
        strerr_die2x(111, FATAL, (char*)malformed);
      gid=l;
      if (gid!=l)
        strerr_die2x(111, FATAL, "group id overflows gid_t");
      groups[ngroups]=gid;
      ++ngroups;
    }
    if (setgroups(ngroups, groups)!=0)
      strerr_die2sys(111, FATAL, "unable to set supplementary groups: ");
  }

  if (do_setuid!=0) {
    s=env_get("UID");
    if (s==0)
      strerr_die2x(111, FATAL, "$UID not set");
    i=scan_ulong(s, &l);
    if (i==0 || s[i]!='\0')
      strerr_die2x(111, FATAL, "$UID not numeric");
    if (l!=(uid_t)l)
      strerr_die2x(111, FATAL, "$UID overflows uid_t");
    if (setuid(l)!=0)
      strerr_die4sys(111, FATAL, "unable to setuid to ", s, ": ");
  }

  if (do_home!=0) {
    s=env_get("HOME");
    if (s==0)
      strerr_die2x(111, FATAL, "$HOME not set");
    if (chdir(s)!=0)
      strerr_die4sys(111, FATAL, "unable to chdir to ", s, ": ");
  }

  {
    char* args[]={ 0, 0 };
    stralloc sa={ 0 };
    if (argc>1) {
      argv++;
      s=argv[0];
    } else {
      s=env_get("SHELL");
      if (s==0)
        strerr_die2x(111, FATAL, "$SHELL not set");
      argv=args;
      argv[0]=s;
    }
    if (do_dash!=0) {
      i=str_rchr(s, '/');
      if (s[i]=='\0')
        i=0;
      else
        ++i;
      if (stralloc_copys(&sa, "-")==0 ||
          stralloc_cats(&sa, s+i)==0 ||
          stralloc_0(&sa)==0)
        strerr_die1sys(111, FATAL);
      argv[0]=sa.s;
    }
    pathexec_run(s, argv, environ);
  }

  strerr_die4sys(111, FATAL, "unable to execute ", s, ": ");
  return 111;
}
